"""
Abstract Object Encoder/Decoder

Object schema is specified in JSON Abstract Data Notation (JADN) format.

Codec currently supports three JSON concrete message formats (verbose,
concise, and minified) but can be extended to support XML or binary formats.

Copyright 2016, 2019 David Kemp
Licensed under the Apache License, Version 2.0
http://www.apache.org/licenses/LICENSE-2.0
"""

import numbers
import re
from jadn.definitions import *
from jadn.utils import topts_s2d, ftopts_s2d, get_config
from jadn.transform import simplify
from jadn.codec.format_validate import format_validators, get_format_validate_function
from jadn.codec.format_serialize_json import json_format_codecs, get_format_encode_function, get_format_decode_function

__version__ = '0.2'

# TODO: add DEFAULT

# Codec Table fields
C_DEC = 0       # Decode function
C_ENC = 1       # Encode function
C_ETYPE = 2     # Encoded type

# Symbol Table fields
S_TDEF = 0      # JADN type definition
S_ENCODE = 1    # Encoder for this type
S_DECODE = 2    # Decoder for this type
S_ENCTYPE = 3   # Encoded value type
S_TOPTS = 4     # Type Options (dict format)
S_FVALIDATE = 5 # Format semantic validation - returns True if valid
S_FENCODE = 6   # Format encode conversion - returns serialized data representation
S_FDECODE = 7   # Format decode conversion - returns API value
S_DMAP = 8      # Decode: Encoded field key or enum value to API
S_EMAP = 9      # Encode: API field key or enum value to Encoded
S_FLD = 10      # Field entries (definition and decoded options)

# Symbol Table Field Definition fields
S_FDEF = 0      # JADN field definition
S_FOPT = 1      # Field Options (dict format)


class Codec:
    """
    Serialize (encode) and De-serialize (decode) values, validate against JADN syntax.

    verbose_rec - True: Record types encoded as maps
                 False: Record types encoded as arrays
    verbose_str - True: Identifiers encoded as strings
                 False: Identifiers encoded as integer tags

    """

    def _error(self, mesg):
        raise ValueError('Validation Error: ' + mesg)

    def __init__(self, schema, verbose_rec=False, verbose_str=False, config=None):
        assert set(enctab) == set(CORE_TYPES)
        self.schema = simplify(schema)             # Convert extensions to core definitions
        conf = config if config else schema
        self.config = get_config(conf['meta'] if 'meta' in conf else None)
        self.format_validate = format_validators()      # Initialize format validation functions
        self.format_codec = json_format_codecs()        # Initialize format serialization functions
        self.types = {t[TypeName]: t for t in self.schema['types']}  # pre-index types to allow symtab forward refs
        self.symtab = None                      # Symbol table - pre-computed values for all datatypes
        self.verbose_rec = None                 # define placeholders in __init__, set value in set_mode
        self.verbose_str = None
        self.set_mode(verbose_rec, verbose_str) # Create symbol table based on encoding mode

    def decode(self, datatype, sval):           # Decode serialized value into API value
        try:
            ts = self.symtab[datatype]
        except KeyError:
            self._error('Decode: datatype "%s" is not defined' % datatype)
        return ts[S_DECODE](ts, sval, self)     # Dispatch to type-specific decoder

    def encode(self, datatype, aval):  # Encode API value into serialized value
        try:
            ts = self.symtab[datatype]
        except KeyError:
            self._error('Encode: datatype "%s" is not defined' % datatype)
        return ts[S_ENCODE](ts, aval, self)     # Dispatch to type-specific encoder

    def set_mode(self, verbose_rec=False, verbose_str=False):
        def symf(fld):      # Build symbol table field entries
            fo, to = ftopts_s2d(fld[FieldOptions])
            if to:
                self._error('%s: internal error: unexpected type options: %s' % (fld[FieldName], str(to)))
            fopts = {'minc': 1, 'maxc': 1}
            fopts.update(fo)
            assert fopts['minc'] in (0, 1) and fopts['maxc'] == 1     # Other cardinalities have been simplified
            fs = [
                fld,        # S_FDEF: JADN field definition
                fopts       # S_FOPT: Field options (dict)
            ]
            return fs

        def make_typeref_pattern(nsid, typename):   # Generate TypeRef pattern - concatenate NSID: and TypeName patterns
            ns = nsid.lstrip('^').rstrip('$')
            tn = typename.lstrip('^').rstrip('$')
            return {'pattern': '^(' + ns + ':)?' + tn + '$'}

        def config_opts(opts):                  # Set configurable option values
            op = [(v[0] + self.config[v[1:]]) if len(v) > 1 and v[1] == '$' else v for v in opts]
            return topts_s2d(op)

        def sym(t):  # Build symbol table based on encoding modes
            symval = [
                t,                              # 0: S_TDEF:  JADN type definition
                enctab[t[BaseType]][C_ENC],     # 1: S_ENCODE: Encoder for this type
                enctab[t[BaseType]][C_DEC],     # 2: S_DECODE: Decoder for this type
                enctab[t[BaseType]][C_ETYPE],   # 3: S_ENCTYPE: Encoded value type
                config_opts(t[TypeOptions]),    # 4: S_TOPTS:  Type Options (dict)
                {},                             # 5: S_FVALIDATE: Format semantic validation - returns True if valid
                {},                             # 6: S_FENCODE: Format encode conversion - returns serialized data representation
                {},                             # 7: S_FDECODE: Format decode conversion - returns API value
                {},                             # 8: S_DMAP: Encoded field key or enum value to API
                {},                             # 9: S_EMAP: API field key or enum value to Encoded
                {},                             # 10: S_FLD: Symbol table field entry
            ]

            if t[BaseType] == 'Record':
                symval[S_ENCODE] = _encode_maprec   # if self.verbose_rec else _encode_array
                symval[S_DECODE] = _decode_maprec   # if self.verbose_rec else _decode_array
                symval[S_ENCTYPE] = dict if self.verbose_rec else list
            fx = FieldName if verbose_str else FieldID
            if t[BaseType] in ('Enumerated', 'Choice', 'Map', 'Record'):
                fx, fa = (FieldID, FieldID) if 'id' in symval[S_TOPTS] else (fx, FieldName)
                try:
                    symval[S_DMAP] = {f[fx]: f[fa] for f in t[Fields]}
                    symval[S_EMAP] = {f[fa]: f[fx] for f in t[Fields]}
                except IndexError:
                    print('symval index error')
                    raise
                if t[BaseType] in ['Choice', 'Map', 'Record']:
                    symval[S_FLD] = {f[fx]: symf(f) for f in t[Fields]}
            elif t[BaseType] == 'Array':
                symval[S_FLD] = {f[FieldID]: symf(f) for f in t[Fields]}
            if t[BaseType] in ('Binary', 'String', 'Array', 'ArrayOf', 'Map', 'MapOf', 'Record'):
                opts = symval[S_TOPTS]
                if 'pattern' not in opts:
                    minv = opts['minv'] if 'minv' in opts else 0
                    maxv = opts['maxv'] if 'maxv' in opts else 0
                    if minv < 0 or maxv < 0:
                        self._error(('%s: length cannot be negative: {}..{}' % t[TypeName], minv, maxv))
                    if maxv == 0:
                        maxv = self.config['$MaxBinary'] if t[BaseType] == 'Binary' else \
                               self.config['$MaxString'] if t[BaseType] == 'String' else \
                               self.config['$MaxElements']
                    opts.update({'minv': max(minv, 0), 'maxv': maxv})
            fmt = symval[S_TOPTS]['format'] if 'format' in symval[S_TOPTS] else ''
            symval[S_FVALIDATE] = get_format_validate_function(self.format_validate, t[BaseType], fmt)
            symval[S_FENCODE] = get_format_encode_function(self.format_codec, t[BaseType], fmt)
            symval[S_FDECODE] = get_format_decode_function(self.format_codec, t[BaseType], fmt)
            return symval

        self.verbose_rec = verbose_rec
        self.verbose_str = verbose_str
        self.symtab = {t[TypeName]: sym(t) for t in self.schema['types']}
        if 'TypeRef' in self.types:
            self.symtab['TypeRef'][S_TOPTS] = make_typeref_pattern(self.config['$NSID'], self.config['$TypeName'])
        self.symtab.update(
            {t: [
                ['', t],
                enctab[t][C_ENC],
                enctab[t][C_DEC],
                enctab[t][C_ETYPE],
                {},
                get_format_validate_function(self.format_validate, t[BaseType], ''),
                get_format_encode_function(self.format_codec, t[BaseType], ''),
                get_format_decode_function(self.format_codec, t[BaseType], ''),
                {},
                {},
                {},
            ] for t in SIMPLE_TYPES})


def _bad_index(ts, k, val):
    td = ts[S_TDEF]
    raise ValueError(
        '%s(%s): array index %d out of bounds (%d, %d)' % (td[TypeName], td[BaseType], k, len(ts[S_FLD]), len(val)))


def _bad_choice(ts, val):
    td = ts[S_TDEF]
    raise ValueError('%s: choice must have one value: %s' % (td[TypeName], val))


def _bad_value(ts, val, fld=None):
    td = ts[S_TDEF]
    if fld is not None:
        raise ValueError('%s(%s): missing required field "%s": %s' % (td[TypeName], td[BaseType], fld[FieldName], val))
    else:
        v = next(iter(val)) if type(val) == dict else val
        raise ValueError('%s(%s): bad value: %s' % (td[TypeName], td[BaseType], v))


def _check_type(ts, val, vtype, fail=False):  # fail forces rejection of boolean vals for number types
    if vtype is not None:
        if fail or not isinstance(val, vtype):
            td = ts[S_TDEF]
            tn = ('%s(%s)' % (td[TypeName], td[BaseType]) if td else 'Primitive')
            raise TypeError('%s: %s is not %s' % (tn, val, vtype))


def _format_encode(ts, val):
    try:
        ts[S_FVALIDATE](val)
    except ValueError:
        raise ValueError('%s: %s is not format "%s"' % (ts[S_TDEF][TypeName], val, ts[S_TOPTS]['format']))
    return ts[S_FENCODE](val)


def _format_decode(ts, val):
    aval = ts[S_FDECODE](val)
    try:
        ts[S_FVALIDATE](aval)
    except ValueError:
        raise ValueError('%s: %s is not format "%s"' % (ts[S_TDEF][TypeName], val, ts[S_TOPTS]['format']))
    return aval


def _check_key(ts, val):
    try:
        return int(val) if isinstance(next(iter(ts[S_DMAP])), int) else val
    except ValueError:
        raise ValueError('%s: %s is not a valid field ID' % (ts[S_TDEF][TypeName], val))


def _check_pattern(ts, val):
    op = ts[S_TOPTS]
    if 'pattern' in op and not re.match(op['pattern'], val):
        tn = ts[S_TDEF][TypeName]
        raise ValueError('%s: string %s does not match %s' % (tn, val, op['pattern']))
    return val


def _check_range(ts, val):
    op = ts[S_TOPTS]
    tn = ts[S_TDEF][TypeName]
    if 'minv' in op and val < op['minv']:
        raise ValueError('%s: %s < minimum %s' % (tn, len(val), op['minv']))
    if 'maxv' in op and val > op['maxv']:
        raise ValueError('%s: %s > maximum %s' % (tn, len(val), op['maxv']))
    return val


def _check_size(ts, val):
    op = ts[S_TOPTS]
    tn = ts[S_TDEF][TypeName]
    if 'minv' in op and len(val) < op['minv']:
        raise ValueError('%s: length %s < minimum %s' % (tn, len(val), op['minv']))
    if 'maxv' in op and len(val) > op['maxv']:
        raise ValueError('%s: length %s > maximum %s' % (tn, len(val), op['maxv']))
    return val


def _extra_value(ts, val, fld):
    td = ts[S_TDEF]
    raise ValueError('%s(%s): unexpected field: %s not in %s:' % (td[TypeName], td[BaseType], val, fld))


def _encode_binary(ts, aval, codec):    # Encode bytes to string
    _check_type(ts, aval, bytes)
    _check_size(ts, aval)
    return _format_encode(ts, aval)


def _decode_binary(ts, sval, codec):    # Decode ASCII string to bytes
    aval = _format_decode(ts, sval)
    _check_type(ts, aval, bytes)        # assert format decode returns correct type
    return _check_size(ts, aval)


def _encode_boolean(ts, val, codec):
    _check_type(ts, val, bool)
    return val


def _decode_boolean(ts, val, codec):
    _check_type(ts, val, bool)
    return val


def _encode_integer(ts, aval, codec):
    _check_type(ts, aval, numbers.Integral, isinstance(aval, bool))
    _check_range(ts, aval)
    return _format_encode(ts, aval)


def _decode_integer(ts, sval, codec):
    aval = _format_decode(ts, sval)
    _check_type(ts, aval, numbers.Integral, isinstance(aval, bool))
    return _check_range(ts, aval)


def _encode_number(ts, aval, codec):
    _check_type(ts, aval, numbers.Real, isinstance(aval, bool))
    _check_range(ts, aval)
    return _format_encode(ts, aval)


def _decode_number(ts, sval, codec):
    aval = _format_decode(ts, sval)
    _check_type(ts, aval, numbers.Real, isinstance(aval, bool))
    return _check_range(ts, aval)


def _encode_null(ts, aval, codec):      # TODO: serialization-specific handling
    _check_type(ts, aval, type(''))
    if aval:
        _bad_value(ts, aval)
    return aval


def _decode_null(ts, val, codec):
    _check_type(ts, val, type(''))
    if val:
        _bad_value(ts, val)
    return val


def _encode_string(ts, aval, codec):
    _check_type(ts, aval, type(''))
    _check_size(ts, aval)
    _check_pattern(ts, aval)
    return _format_encode(ts, aval)


def _decode_string(ts, sval, codec):
    aval = _format_decode(ts, sval)
    _check_type(ts, aval, type(''))
    _check_size(ts, aval)
    return _check_pattern(ts, aval)


def _encode_enumerated(ts, aval, codec):                # TODO: Serialization
    _check_type(ts, aval, type(next(iter(ts[S_EMAP]))))
    if aval in ts[S_EMAP]:
        return ts[S_EMAP][aval]
    else:
        td = ts[S_TDEF]
        raise ValueError('%s: %r is not a valid %s' % (td[BaseType], aval, td[TypeName]))


def _decode_enumerated(ts, sval, codec):
    _check_type(ts, sval, type(next(iter(ts[S_DMAP]))))
    if sval in ts[S_DMAP]:
        return ts[S_DMAP][sval]
    else:
        td = ts[S_TDEF]
        raise ValueError('%s: %r is not a valid %s' % (td[BaseType], sval, td[TypeName]))


def _encode_choice(ts, val, codec):
    _check_type(ts, val, dict)
    if len(val) != 1:
        _bad_choice(ts, val)
    k, v = next(iter(val.items()))
    if k not in ts[S_EMAP]:
        _bad_value(ts, val)
    k = ts[S_EMAP][k]
    f = ts[S_FLD][k][S_FDEF]
    return {k: codec.encode(f[FieldType], v)}


def _decode_choice(ts, val, codec):  # Map Choice:  val == {key: value}
    _check_type(ts, val, dict)
    if len(val) != 1:
        _bad_choice(ts, val)
    k, v = next(iter(val.items()))
    k = _check_key(ts, k)
    if k not in ts[S_DMAP]:
        _bad_value(ts, val)
    f = ts[S_FLD][k][S_FDEF]
    k = ts[S_DMAP][k]
    return {k: codec.decode(f[FieldType], v)}


def _encode_maprec(ts, aval, codec):
    _check_type(ts, aval, dict)
    sval = ts[S_ENCTYPE]()
    assert type(sval) in (list, dict)
    fx = FieldName if codec.verbose_str else FieldID  # Verbose or minified identifier strings
    fnames = [f[S_FDEF][FieldName] for f in ts[S_FLD].values()]
    for f in ts[S_TDEF][Fields]:
        fs = ts[S_FLD][f[fx]]  # Symtab entry for field
        fd = fs[S_FDEF]  # JADN field definition from symtab
        fname = fd[FieldName]  # Field name
        fopts = fs[S_FOPT]  # Field options dict
        if 'tfield' in fopts:  # Type of this field is specified by contents of another field
            choice_type = aval[fopts['tfield']]
            e = codec.encode(fd[FieldType], {choice_type: aval[fname]})
            sv = next(iter(e.values()))
        else:
            sv = codec.encode(fd[FieldType], aval[fname]) if fname in aval else None
        if sv is None and ('minc' not in fopts or fopts['minc'] > 0):  # Missing required field
            _bad_value(ts, aval, fd)
        if type(sval) == list:  # Concise Record
            sval.append(sv)
        elif sv is not None:  # Map or Verbose Record
            sval[fd[fx]] = sv

    if set(aval) - set(fnames):
        _extra_value(ts, aval, fnames)
    if type(sval) == list:
        while sval and sval[-1] is None:  # Strip non-populated trailing optional values
            sval.pop()
    return sval


def _decode_maprec(ts, sval, codec):
    _check_type(ts, sval, ts[S_ENCTYPE])
    val = sval
    if ts[S_ENCTYPE] == dict:
        val = {_check_key(ts, k): v for k, v in sval.items()}
    aval = dict()
    fx = FieldName if codec.verbose_str else FieldID  # Verbose or minified identifier strings
    fnames = [k for k in ts[S_FLD]]
    for f in ts[S_TDEF][Fields]:
        fs = ts[S_FLD][f[fx]]  # Symtab entry for field
        fd = fs[S_FDEF]  # JADN field definition from symtab
        fopts = fs[S_FOPT]  # Field options dict
        if type(val) == dict:
            fn = f[fx]
            sv = val[fn] if fn in val else None
        else:
            fn = fd[FieldID] - 1
            sv = val[fn] if len(val) > fn else None
        if sv is not None:
            if 'tfield' in fopts:  # Type of this field is specified by contents of another field
                ctf = fopts['tfield']
                choice_type = val[ctf] if isinstance(val, dict) else val[ts[S_EMAP][ctf] - 1]
                av = codec.decode(fd[FieldType], {choice_type: sv})
                aval[fd[FieldName]] = next(iter(av.values()))
            else:
                aval[fd[FieldName]] = codec.decode(fd[FieldType], sv)
        else:
            if 'minc' not in fopts or fopts['minc'] > 0:
                _bad_value(ts, val, fd)
    extra = set(val) - set(fnames) if type(val) == dict else len(val) > len(ts[S_FLD])
    if extra:
        _extra_value(ts, val, extra)
    return aval


def _encode_array(ts, aval, codec):
    _check_type(ts, aval, list)
    sval = list()
    extra = len(aval) > len(ts[S_FLD])
    if extra:
        _extra_value(ts, aval, extra)
    for fn in ts[S_TDEF][Fields]:
        f = ts[S_FLD][fn[FieldID]][S_FDEF]  # Use symtab field definition
        fx = f[FieldID] - 1
        fopts = ts[S_FLD][fx + 1][S_FOPT]
        av = aval[fx] if len(aval) > fx else None
        if av is not None:
            if 'tfield' in fopts:
                choice_type = aval[int(fopts['tfield']) - 1]
                e = codec.encode(f[FieldType], {choice_type: av})
                sv = e[next(iter(e))]
            else:
                sv = codec.encode(f[FieldType], av)
            sval.append(sv)
        else:
            sval.append(None)
            if 'minc' in fopts and fopts['minc'] > 0:   # Value is required
                _bad_value(ts, aval, f)
    while sval and sval[-1] is None:            # Strip non-populated trailing optional values
        sval.pop()
    return _format_encode(ts, sval)


def _decode_array(ts, sval, codec):  # Ordered list of types, returned as a list
    val = _format_decode(ts, sval)
    _check_type(ts, val, list)
    aval = list()
    extra = len(val) > len(ts[S_FLD])
    if extra:
        _extra_value(ts, val, extra)  # TODO: write sensible display of excess values
    for fn in ts[S_TDEF][Fields]:
        f = ts[S_FLD][fn[FieldID]][S_FDEF]  # Use symtab field definition
        fx = f[FieldID] - 1
        fopts = ts[S_FLD][fx + 1][S_FOPT]
        sv = val[fx] if len(val) > fx else None
        if sv is not None:
            if 'tfield' in fopts:
                choice_type = val[int(fopts['tfield']) - 1]
                d = codec.decode(f[FieldType], {choice_type: sv})  # TODO: fix str/int handling of choice
                av = d[next(iter(d))]
            else:
                av = codec.decode(f[FieldType], sv)
            aval.append(av)
        else:
            aval.append(None)
            if 'minc' not in fopts or fopts['minc'] > 0:
                _bad_value(ts, val, f)
    while aval and aval[-1] is None:  # Strip non-populated trailing optional values
        aval.pop()
    return aval


def _encode_array_of(ts, val, codec):
    _check_type(ts, val, list)
    _check_size(ts, val)
    return [codec.encode(ts[S_TOPTS]['vtype'], v) for v in val]


def _decode_array_of(ts, val, codec):
    _check_type(ts, val, list)
    _check_size(ts, val)
    return [codec.decode(ts[S_TOPTS]['vtype'], v) for v in val]


def _encode_map_of(ts, aval, codec):
    _check_type(ts, aval, dict)
    _check_size(ts, aval)
    to = ts[S_TOPTS]
    return {codec.encode(to['ktype'], k): codec.encode(to['vtype'], v) for k, v in aval.items()}


def _decode_map_of(ts, sval, codec):
    _check_type(ts, sval, dict)
    _check_size(ts, sval)
    return {k: codec.decode(ts[S_TOPTS]['vtype'], v) for k, v in sval.items()}


enctab = {  # decode, encode, min encoded type
    'Binary': (_decode_binary, _encode_binary, str),
    'Boolean': (_decode_boolean, _encode_boolean, bool),
    'Integer': (_decode_integer, _encode_integer, int),
    'Number': (_decode_number, _encode_number, float),
    'Null': (_decode_null, _encode_null, str),
    'String': (_decode_string, _encode_string, str),
    'Enumerated': (_decode_enumerated, _encode_enumerated, int),
    'Choice': (_decode_choice, _encode_choice, dict),
    'Array': (_decode_array, _encode_array, list),
    'ArrayOf': (_decode_array_of, _encode_array_of, list),
    'Map': (_decode_maprec, _encode_maprec, dict),
    'MapOf': (_decode_map_of, _encode_map_of, dict),
    'Record': (None, None, None),  # Dynamic values
}



"""
Translate JADN to JSON Schema
"""

import json
from jadn.definitions import *
from jadn.utils import topts_s2d, ftopts_s2d, get_config
from jadn.transform.transform import get_enum_items
from datetime import datetime


def dmerge(*dicts):             # Merge any number of dicts
    rv = {}
    for d in dicts:
        rv.update(d)
    return rv


def get_items(stype, ctx):       # return enumerated items if stype is Enumerated
    td = ctx['type_defs']
    et = stype[1:] if stype[0] == OPTION_ID['enum'] else stype
    if et in td and td[et][BaseType] == 'Enumerated':
        to = topts_s2d(td[et][TypeOptions])
        et = to['enum'] if 'enum' in to else et
    if et in td and \
        td[et][BaseType] in ('Enumerated', 'Array', 'Choice', 'Map', 'Record') and \
        'id' not in topts_s2d(td[et][TypeOptions]):
        return [f[ItemValue] for f in td[et][Fields]]


def spaces(s):                  # Return name with dash and underscore replaced with space
    return s.replace('-', ' ').replace('_', ' ')


def fieldname(name):            # Return TypeName converted to FieldName
    return name.lower().replace('-', '_')

# === Return JSON Schema keywords


def w_td(tname, desc):              # Make type and type description
    return {'type': tname, 'description': desc} if desc else {'type': tname}


def w_def(typ):                     # Make definition reference
    return {'#ref': '#/definitions/' + typ}


def w_kvtype(stype, ctx):            # Make definition from ktype or vtype option
    stypes = {'Boolean': 'boolean', 'Integer': 'integer', 'Number': 'number', 'String': 'string'}
    if stype in stypes:
        return {'type': stypes[stype]}
    if stype[0] in (OPTION_ID['enum'], OPTION_ID['pointer']):
        tdef = ctx['type_defs'][stype[1:]]
        topts = topts_s2d(tdef[TypeOptions])
        fields = get_enum_items(tdef, topts, ctx['type_defs'])
        idopt = 'id' in topts
        return w_enum(fields, FieldID if idopt else FieldName, FieldDesc, idopt, ctx)
    return {'$ref': '#/definitions/' + stype}


def w_enum(fields, vcol, dcol, idopt, ctx):
    def pattern(vals):              # Return an enum regex from a list of values
        return '^(' + '|'.join([str(v) for v in vals]) + ')$'

    def fdesc(fld, desc, idopt):    # Make field description
        return fld[FieldName] + ' - ' + desc if idopt else desc

    values = [f[vcol] for f in fields]
    es = ctx['enum_style']
    assert es in ['enum', 'const', 'regex']
    if es == 'regex':
        return {'type': 'string', 'pattern': pattern(values)}
    elif es == 'const':
        return {'oneOf': [{'const': f[vcol], 'description': fdesc(f, f[dcol], idopt)} for f in fields]}
    else:
        return {'enum': values}

def w_export(exp, ctx):          # Make top-level definition header
    if len(exp) == 1:
        return {'$ref': '#/definitions/' + exp[0]}
    return {
        'type': 'object',
        'additionalProperties': False,
        'properties': {fieldname(ctx['type_defs'][t][TypeName]): {'$ref': '#/definitions/' + t} for t in exp}
    }

def w_ref(tname, ctx):
    nsid, tn = tname.split(':', maxsplit=1) if ':' in tname else [None, tname]
    assert not is_builtin(tn)
    if nsid:
        imp = {'$ref': ctx['meta_imps'][nsid] + '/definitions/' + tn} if ctx['import_style'] == 'ref' else {}
        ctx['imported_types'][nsid].update({tn: imp})
        return {'$ref': '#/imports/' + nsid + '/' + tn}
    return {'$ref': '#/definitions/' + tn}


def w_fdef(f, ctx):                      # Make field definition
    fopts, topts = ftopts_s2d(f[FieldOptions])
    if is_builtin(f[FieldType]):
        t = w_type(['', f[FieldType], [], f[FieldDesc]], topts, ctx)
    else:
        t = dmerge(w_ref(f[FieldType], ctx), {'description': f[FieldDesc]})

    minv = max(fopts['minc'], 1) if 'minc' in fopts else 1
    maxv = fopts['maxc'] if 'maxc' in fopts else 1
    return t if minv <= 1 and maxv == 1 else dmerge(
        {'type': 'array'},
        {'description': f[FieldDesc]},
        {'uniqueItems': True} if 'unique' in topts else {},
        {'minItems': minv} if minv != 0 else {},
        {'maxItems': maxv} if maxv != 0 else {},
        {'items': t}
    )


def w_format(fmt):                  # Make semantic validation keywords
    jadn_fmt = {
        'x': {'contentEncoding': 'base16'},
        # 'eui': {'pattern': r'^([0-9a-fA-F]{2}[:-]){5}[0-9a-fA-F]{2}$'},
        'eui': {'pattern': r'^([0-9a-fA-F]{2}[:-]){5}[0-9A-Fa-f]{2}(([:-][0-9A-Fa-f]{2}){2})?$'},
        'ipv4-addr': {'pattern': r'^((25[0-5]|2[0-4][0-9]|[01]?[0-9]?[0-9])\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9]?[0-9])$'},
        'ipv6-addr': {'pattern': r'^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))(%.+)$'},
        'ipv4-net': {'pattern': r'^((25[0-5]|2[0-4][0-9]|[01]?[0-9]?[0-9])\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9]?[0-9])(\/(3[0-2]|[0-2]?[0-9]))?$'},
        'ipv6-net': {'pattern': r'^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))(%.+)?s*(\/([0-9]|[1-9][0-9]|1[0-1][0-9]|12[0-8]))?$'},
        'i8': {'minimum': -128, 'maximum': 127},
        'i16': {'minimum': -32768, 'maximum': 32767},
        'i32': {'minimum': -2147483648, 'maximum': 2147483647}
    }
    return (jadn_fmt[fmt] if fmt in jadn_fmt else {'format': fmt}) if fmt else {}

# ========== JADN Types ==========:


def t_binary(tdef, topts, ctx):
    return dmerge(
        w_td('string', tdef[TypeDesc]),
        w_format(topts['format']) if 'format' in topts else {'contentEncoding': 'base64url'},

        # TODO: Fixme: JSON Schema cannot express length of content-encoded data
        # This would require adjusting string length by 2 for hex and 4/3 for base64
        # Impossible to calculate string length for other string formats
        # {'minLength': topts['minv']} if 'minv' in topts and topts['minv'] > 0 else {},
        # {'maxLength': topts['maxv']} if 'maxv' in topts else {}
    )


def t_boolean(tdef, topts, ctx):
    return w_td('boolean', tdef[TypeDesc])


def t_integer(tdef, topts, ctx):
    return dmerge(
        w_td('integer', tdef[TypeDesc]),
        {'minimum': topts['minv']} if 'minv' in topts else {},
        {'maximum': topts['maxv']} if 'maxv' in topts else {}
    )


def t_number(tdef, topts, ctx):
    return dmerge(
        w_td('number', tdef[TypeDesc]),
        {'minimum': topts['minv']} if 'minv' in topts else {},
        {'maximum': topts['maxv']} if 'maxv' in topts else {}
    )


def t_null(tdef, topts, ctx):
    return {}


def t_string(tdef, topts, ctx):
    return dmerge(
        w_td('string', tdef[TypeDesc]),
        w_format(topts['format']) if 'format' in topts else {},
        {'minLength': topts['minv']} if 'minv' in topts and topts['minv'] != 0 else {},
        {'maxLength': topts['maxv']} if 'maxv' in topts and topts['maxv'] != 0 else {},
        {'pattern': topts['pattern']} if 'pattern' in topts else {}
    )


def t_enumerated(tdef, topts, ctx):
    item, tname = (ItemID, 'integer') if 'id' in topts else (ItemValue, 'string')
    fields = get_enum_items(tdef, topts, ctx['type_defs'])
    dcol = FieldDesc if 'enum' in topts or 'pointer' in topts else ItemDesc
    return dmerge(
        w_td(tname, tdef[TypeDesc]),
        w_enum(fields, item, dcol, 'id' in topts, ctx)
    )


def t_choice(tdef, topts, ctx):
    return dmerge(
        w_td('object', tdef[TypeDesc]),
        {'additionalProperties': False, 'minProperties': 1, 'maxProperties': 1},
        {'properties': {f[FieldName]: w_fdef(f, ctx) for f in tdef[Fields]}}
    )


def t_array(tdef, topts, ctx):
    fmt = w_format(topts['format']) if 'format' in topts else {}
    if fmt:
        return dmerge(w_td('string', tdef[TypeDesc]), fmt)
    return dmerge(
        w_td('array', tdef[TypeDesc]),
        {'additionalItems': False},
        {'minItems': topts['minv']} if 'minv' in topts and topts['minv'] != 0 else {},
        {'maxItems': topts['maxv']} if 'maxv' in topts and topts['maxv'] != 0 else {},
        {'items': [w_fdef(f, ctx) for f in tdef[Fields]]}
    )


def t_array_of(tdef, topts, ctx):
    return dmerge(
        w_td('array', tdef[TypeDesc]),
        {'uniqueItems': True} if 'unique' in topts else {},
        {'minItems': topts['minv']} if 'minv' in topts and topts['minv'] != 0 else {},
        {'maxItems': topts['maxv']} if 'maxv' in topts and topts['maxv'] != 0 else {},
        {'items': w_kvtype(topts['vtype'], ctx)}
    )


def t_map(tdef, topts, ctx):
    def req(f):
        fo, to = ftopts_s2d(f[FieldOptions])
        return fo['minc'] >= 1 if 'minc' in fo else True
    required = [f[FieldName] for f in tdef[Fields] if req(f)]
    return dmerge(
        w_td('object', tdef[TypeDesc]),
        {'additionalProperties': False},
        {'required': required} if required else {},
        {'minProperties': topts['minv']} if 'minv' in topts and topts['minv'] != 0 else {},
        {'maxProperties': topts['maxv']} if 'maxv' in topts and topts['maxv'] != 0 else {},
        {'properties': {f[FieldName]: w_fdef(f, ctx) for f in tdef[Fields]}}
    )


def t_map_of(tdef, topts, ctx):
    items = get_items(topts['ktype'], ctx)
    vtype = w_kvtype(topts['vtype'], ctx)
    return dmerge(
        w_td('object', tdef[TypeDesc]),
        {'additionalProperties': False},
        {'minProperties': topts['minv']} if 'minv' in topts and topts['minv'] != 0 else {},
        {'maxProperties': topts['maxv']} if 'maxv' in topts and topts['maxv'] != 0 else {},
        {'patternProperties': {pattern(items): vtype}} if items and ctx['enum_style'] == 'regex' else
        {'properties': {f: vtype for f in items}} if items else {}
    )


def w_type(tdef, topts, ctx):       # Write a JADN type definition in JSON Schema format
    config_max = {
        'Binary':  '$MaxBinary',
        'String':  '$MaxString',
        'Array':   '$MaxElements',
        'ArrayOf': '$MaxElements',
        'Map':     '$MaxElements',
        'MapOf':   '$MaxElements',
        'Record':  '$MaxElements'
    }
    wtype = {
        'Binary': t_binary,
        'Boolean': t_boolean,
        'Integer': t_integer,
        'Number': t_number,
        'Null': t_null,
        'String': t_string,
        'Enumerated': t_enumerated,
        'Choice': t_choice,
        'Array': t_array,
        'ArrayOf': t_array_of,
        'Map': t_map,
        'MapOf': t_map_of,
        'Record': t_map if ctx['verbose'] else t_array
    }
    if 'maxv' in topts and topts['maxv'] == 0:
        topts['maxv'] = ctx['config'][config_max[tdef[BaseType]]]
    sc = wtype[tdef[BaseType]](tdef, topts, ctx)
    if 'and' in topts:
        return {'allOf': [sc, w_ref(topts['and'], ctx)]}
    if 'or' in topts:
        return {'anyOf': [sc, w_ref(topts['or'], ctx)]}
    return sc

# ========== Make JSON Schema ==========


def json_schema_dumps(jadn, verbose=True, enum_style='enum', import_style='any'):  # TODO: use string IDs if verbose=False
    #  verbose True: Record->object, field Names;  False: Record->array, field IDs
    assert enum_style in ('const', 'enum', 'regex')
    #  const: generate oneOf keyword with const for each item
    #  enum: generate an enum keyword containing all items
    #  regex: generate a regular expression that matches each item
    assert import_style in ('any', 'ref')
    #  any: ignore types defined in other modules, validate anything
    #  ref: generate $ref keywords that must be resolved before the JSON Schema can validate referenced types

    meta = jadn['meta'] if 'meta' in jadn else {}
    td = {t[TypeName]: t for t in jadn['types']}    # Build index of type definitions
    exports = [e for e in meta['exports'] if e in td] if 'exports' in meta else []
    imported_types = {k: {} for k in meta['imports']} if 'imports' in meta else {}
    ctx = {                                         # Translation context
        'config': get_config(meta),
        'type_defs': td,
        'verbose': verbose,
        'enum_style': enum_style,
        'imported_types': imported_types,
        'import_style': import_style,
        'meta_imps': meta['imports'] if 'imports' in meta else None
    }

    def tt(tdef, ctx):                              # Return type definition with title
        topts = topts_s2d(tdef[TypeOptions])
        return dmerge({'title': spaces(tdef[TypeName])}, w_type(tdef, topts, ctx))

    return json.dumps(dmerge(
        {'$schema': 'http://json-schema.org/draft-07/schema#'},
        {'$id': meta['module'] if 'module' in meta else ''},
        {'title': meta['title'] if 'title' in meta else ''},
        {'description': meta['description']} if 'description' in meta else {},
        w_export(exports, ctx),
        {'definitions': {t: tt(td[t], ctx) for t in (exports + [t for t in td if t not in exports])}},
        {'imports': imported_types} if imported_types else {}
    ), indent=2)


def json_schema_dump(jadn, fname, source=None, verbose=True, enum_style='enum', import_style='ref'):
    with open(fname, 'w') as f:
        if source:
            f.write('/* Generated from ' + source + ', ' + datetime.ctime(datetime.now()) + ' */\n\n')
        f.write(json_schema_dumps(jadn, verbose, enum_style, import_style))



"""
Translate JADN to JAS (JADN Abstract Syntax)
"""

from jadn.definitions import *
from jadn.utils import topts_s2d, ftopts_s2d
from copy import deepcopy
from datetime import datetime
from textwrap import fill

stype_map = {                   # Map JADN built-in types to JAS type names (Equivalent ASN.1 types in comments)
    'Binary': 'BINARY',         # OCTET STRING
    'Boolean': 'BOOLEAN',       # BOOLEAN
    'Integer': 'INTEGER',       # INTEGER
    'Number': 'REAL',           # REAL
    'Null': 'NULL',             # NULL
    'String': 'STRING',         # UTF8String
    'Array': 'ARRAY',           # SEQUENCE
    'ArrayOf': 'ARRAY_OF',      # SEQUENCE OF
    'Choice': 'CHOICE',         # CHOICE
    'Enumerated': 'ENUMERATED', # ENUMERATED
    'Map': 'MAP',               # SET
    'MapOf': 'MAP_OF',          #
    'Record': 'RECORD'          # SEQUENCE
}


def stype(jtype):
    return stype_map[jtype] if jtype in stype_map else jtype


def jas_dumps(jadn):
    """
    Produce JAS module from JADN structure

    JAS represents features available in both JADN and ASN.1 using ASN.1 syntax, but adds
    extended datatypes (Record, Map) for JADN types not directly representable in ASN.1.
    With appropriate encoding rules (which do not yet exist), SEQUENCE could replace Record.
    Map could be implemented using ASN.1 table constraints, but for the purpose of representing
    JSON objects, the Map first-class type in JAS is easier to use.
    """

    jas = '/*\n'
    hdrs = jadn['meta']
    hdr_list = ['module', 'patch', 'title', 'description', 'imports', 'exports', 'bounds']
    for h in hdr_list + list(set(hdrs) - set(hdr_list)):
        if h in hdrs:
            if h == 'description':
                jas += fill(hdrs[h], width=80, initial_indent='{0:14} '.format(h+':'), subsequent_indent=15*' ') + '\n'
            elif h == 'imports':
                hh = '{:14} '.format(h+':')
                for imp in hdrs[h]:
                    jas += hh + '{}: {}\n'.format(*imp)
                    hh = 15*' '
            elif h == 'exports':
                jas += '{:14} {}\n'.format(h+':', ', '.join(hdrs[h]))
            else:
                jas += '{:14} {}\n'.format(h+':', hdrs[h])
    jas += '*/\n'

    assert set(stype_map) == set(CORE_TYPES)         # Ensure type list is up to date
    tolist = ['id', 'vtype', 'ktype', 'enum', 'pointer', 'format', 'pattern', 'minv', 'maxv', 'unique', 'and', 'or']
    assert set(TYPE_OPTIONS.values()) == set(tolist)                # Ensure type options list is up to date
    folist = ['minc', 'maxc', 'tfield', 'dir', 'default']
    assert set(FIELD_OPTIONS.values()) == set(folist)               # Ensure field options list is up to date
    for td in jadn['types']:                    # 0:type name, 1:base type, 2:type opts, 3:type desc, 4:fields
        tname = td[TypeName]
        ttype = td[BaseType]
        topts = topts_s2d(td[TypeOptions])
        tostr = ''
        if 'minv' in topts or 'maxv' in topts:
            lo = topts['minv'] if 'minv' in topts else 0
            hi = topts['maxv'] if 'maxv' in topts else 0
            range = ''
            if lo or hi:
                range = '(' + str(lo) + '..' + (str(hi) if hi else 'MAX') + ')'
        for opt in tolist:
            if opt in topts:
                ov = topts[opt]
                if opt == 'id':
                    tostr += '.ID'
                elif opt =='vtype':
                    tostr += '(' + ov + ')'
                elif opt == 'ktype':
                    pass            # fix MapOf(ktype, vtype)
                elif opt == 'pattern':
                    tostr += ' (PATTERN ("' + ov + '"))'
                elif opt == 'format':
                    tostr += ' (CONSTRAINED BY {' + ov + '})'
                elif opt in ('minv', 'maxv'):     # TODO fix to handle both
                    if range:
                        if ttype in ('Integer', 'Number'):
                            tostr += ' ' + range
                        elif ttype in ('Binary', 'String', 'Array', 'ArrayOf', 'Map', 'MapOf', 'Record'):
                            tostr += ' (Size ' + range + ')'
                        else:
                            assert False        # Should never get here
                    range = ''
                else:
                    tostr += ' %' + opt + ': ' + str(ov) + '%'
        tdesc = '    -- ' + td[TypeDesc] if td[TypeDesc] else ''
        jas += '\n' + tname + ' ::= ' + stype(ttype) + tostr
        if len(td) > Fields:
            titems = deepcopy(td[Fields])
            for n, i in enumerate(titems):      # 0:tag, 1:enum item name, 2:enum item desc  (enumerated), or
                if len(i) > FieldOptions:              # 0:tag, 1:field name, 2:field type, 3: field opts, 4:field desc
                    desc = i[FieldDesc]
                    i[FieldType] = stype(i[FieldType])
                else:
                    desc = i[ItemDesc]
                desc = '    -- ' + desc if desc else ''
                i.append(',' + desc if n < len(titems) - 1 else (' ' + desc if desc else ''))   # TODO: fix hacked desc for join
            flen = min(32, max(12, max([len(i[FieldName]) for i in titems]) + 1 if titems else 0))
            jas += ' {' + tdesc + '\n'
            if ttype.lower() == 'enumerated':
                fmt = '    {1:' + str(flen) + '} ({0:d}){3}'
                jas += '\n'.join([fmt.format(*i) for i in titems])
            else:
                fmt = '    {1:' + str(flen) + '} [{0:d}] {2}{3}{4}'
                if ttype.lower() == 'record':
                    fmt = '    {1:' + str(flen) + '} {2}{3}{4}'
                items = []
                for n, i in enumerate(titems):
                    ostr = ''
                    opts, ftopts = ftopts_s2d(i[FieldOptions])
                    if 'tfield' in opts:
                        ostr += '.&' + opts['tfield']
                        del opts['tfield']
                    if 'vtype' in opts:
                        ostr += '.*'
                        del opts['vtype']
                    if 'minc' in opts:
                        if opts['minc'] == 0:         # TODO: handle array fields (max != 1)
                            ostr += ' OPTIONAL'
                        del opts['minc']
                    items += [fmt.format(i[FieldID], i[FieldName], i[FieldType], ostr, i[5]) + (' %' + str(opts) if opts else '')]
                jas += '\n'.join(items)
            jas += '\n}\n' if titems else '}\n'
        else:
            jas += tdesc + '\n'
    return jas


def jas_dump(jadn, fname, source=''):
    with open(fname, 'w') as f:
        if source:
            f.write('-- Generated from ' + source + ', ' + datetime.ctime(datetime.now()) + '\n\n')
        f.write(jas_dumps(jadn))



"""
Translate JADN to JADN Interface Definition Language
"""

from datetime import datetime

from jadn.definitions import *
from jadn.utils import topts_s2d, ftopts_s2d, typestring


def jidl_dumps(jadn):

    def _mult(opts, optional=False):
        lo = opts.pop('minc', 1)
        hi = opts.pop('maxc', 1)
        if hi == 1:
            return '' if lo == 1 else (' optional' if optional else ' [0..1]')
        h = '*' if hi == 0 else str(hi)
        return ' [' + str(lo) + '..' + str(h) + ']'

    def _fieldstr(typestr, opts):
        if 'minc' in opts or 'maxc' in opts:
            return typestr + _mult(opts, optional=True)
        if 'tfield' in opts:
            return typestr + '(&' + opts['tfield'] + ')'
        return typestr

    def line(cw, content, desc):
        fmt = '{:' + str(cw) + '}{}'
        return fmt.format(content.rstrip(), ' // ' + desc if desc else '').rstrip() + '\n'

    meta = jadn['meta'] if 'meta' in jadn else {}
    text = ''
    meta_list = ('title', 'module', 'patch', 'description', 'exports', 'imports')
    for h in meta_list + tuple(set(meta) - set(meta_list)):
        if h in meta:
            text += '{:>12}: {}\n'.format(h, meta[h])

    efmt = ['  {0:4d} {1:15}',          # Enumerated
            '  {0:4d}']                 # Enumerated.ID
    ffmt = ['  {0:4d} {1:15} {2:20}',   # Full
            '  {0:4d} {2:20}']          # Full.ID

    for td in jadn['types']:
        bt = td[BaseType]
        assert is_builtin(bt)
        to = topts_s2d(td[TypeOptions])
        ts = typestring(bt, to)
        flds = '{' if has_fields(bt) or (bt == 'Enumerated' and 'enum' not in to and 'pointer' not in to) else ''
        text += '\n' + line(44, '{} = {} {}'.format(td[TypeName], ts, flds), td[TypeDesc])
        if flds:
            id = 1 if 'id' in to or bt == 'Array' else 0
            for n, f in enumerate(td[Fields]):
                sep = ',' if n < len(td[Fields]) - 1 else ''
                if bt == 'Enumerated':
                    content = efmt[id].format(f[ItemID], f[ItemValue] + sep)
                    desc = (f[ItemValue] + ':: ' if id and f[ItemValue] else '') + f[ItemDesc]
                    text += line(48, content, desc)
                else:
                    ft, fto = ftopts_s2d(f[FieldOptions])
                    fo = {'minc': 1, 'maxc': 1}
                    fo.update(ft)
                    fs = _fieldstr(typestring(f[FieldType], fto), fo) + (' unique' if 'unique' in fto else '')
                    fn = f[FieldName] + ('/' if 'dir' in fo else '')
                    content = ffmt[id].format(f[FieldID], fn, fs + sep)
                    desc = (f[FieldName] + ':: ' if id and f[FieldName] else '') + f[FieldDesc]
                    text += line(48, content, desc)

        text += '}\n' if flds else ''

    return text


def jidl_dump(jadn, fname, source=''):
    with open(fname, 'w', encoding='utf8') as f:
        if source:
            f.write('/* Generated from ' + source + ', ' + datetime.ctime(datetime.now()) + ' */\n\n')
        f.write(jidl_dumps(jadn))



"""
Translate JADN to HTML or Markdown property tables
"""

from jadn.definitions import *
from jadn.utils import multiplicity, topts_s2d, ftopts_s2d, typestring, get_config
from datetime import datetime


def _fmt(s, f):
    f1 = {'n': '', 's': '', 'd': '', 'b': '**', 'h': '**_'}
    f2 = {'n': '', 's': '', 'd': '', 'b': '**', 'h': '_**'}
    ss = '\*' if s == '*' else s
    return f1[f] + ss + f2[f]

#--------- Markdown ouput -----------------


def doc_begin_m(title):
    text = '## Schema\n'
    return text


def doc_end_m():
    return ''


def sect_m(num, name):
    n = ''
    # n = '.'.join([str(n) for n in num]) + ' '
    # return '\n' + len(num)*'#' + ' ' + n + name + '\n'
    return ''


def meta_begin_m():
    return '| . | . |\n| ---: | :--- |\n'


def meta_item_m(h, val):
    if h == 'exports':
        sval = ', '.join(val)
    elif h in ('imports', 'config'):
        sval = ' '.join(['**' + k + '**:&nbsp;' + str(v).replace('|', '&vert;') for k, v in val.items()])
    else:
        sval = val
    return '| **' + h + ':** | ' + sval + ' |\n'


def meta_end_m():
    return ''


def type_begin_m(tname, ttype, headers, cls):
    assert len(headers) == len(cls)
    ch = {'n': '---:', 'h': '---:', 's': ':---', 'd': ':---', 'b': ':---'}
    clh = [ch[c] if c in ch else '---' for c in cls]
    to = ' (' + ttype + ')' if ttype else ''
    tc = '\n**_Type: ' + tname + to + '_**' if tname else ''
    return tc + '\n\n| ' + ' | '.join(headers) + ' |\n| ' + ' | '.join(clh) + ' |\n'


def type_item_m(row, cls):
    assert len(row) == len(cls)
    return '| ' + ' | '.join([_fmt(*r) for r in zip(row, cls)]) + ' |\n'


def type_end_m():
    return ''


# ---------- HTML output ------------------


def doc_begin_h(title):
    text = '<!DOCTYPE html>\n<html lang="en">\n<head>\n<meta charset="UTF-8">\n'
    text += '<link rel="stylesheet" type="text/css" href="theme.css">\n'
    text += '<title>' + title + '</title>\n</head>\n'
    text += '<body>\n<h2>Schema</h2>\n'
    return text


def doc_end_h():
    return '</body>\n'


def sect_h(num, name):
    hn = 'h' + str(len(num))
    n = ''
    # n = '.'.join([str(n) for n in num]) + ' '
    return '<' + hn + '>' + n + name + '</' + hn + '>\n'


def meta_begin_h():
    return '<table>\n'


def meta_item_h(h, val):
    if h == "exports":
        sval = ', '.join(val)
    elif h in ('imports', 'config'):
        sval = '<br>\n'.join(['<span class="b">' + k + '</span>:&nbsp;' + str(v) for k, v in val.items()])
    else:
        sval = val
    rc = [[h + ':', 'h'], [sval, 's']]
    return '<tr>' + ''.join(['<td class="' + c[1] + '">' + c[0] + '</td>' for c in rc]) + '</tr>\n'


def meta_imps_h(imports):
    return '<br>\n'.join([i[0] + ': ' + i[1] for i in imports])


def meta_end_h():
    return '</table>\n'


def type_begin_h(tname, ttype, headers, cls):
    assert len(headers) == len(cls)
    to = ' (' + ttype + ')' if ttype else ''
    tc = '<caption>' + tname + to + '</caption>' if tname else ''
    rc = zip(headers, cls)
    return '<table>' + tc + '<tr>' + ''.join(['<th class="' + c[1] + '">' + c[0] + '</th>' for c in rc]) + '</tr>\n'


def type_item_h(row, cls):
    assert len(row) == len(cls)
    rc = zip(row, cls)
    return '<tr>' + ''.join(['<td class="' + c[1] + '">' + c[0] + '</td>' for c in rc]) + '</tr>\n'


def type_end_h():
    return '</table>\n'


# ---------- JADN Source (JAS) output ------------------


def doc_begin_s(title):
    text = ''
    return text


def doc_end_s():
    return ''


def sect_s(num, name):
    return ''


def meta_begin_s():
    return ''


def meta_item_s(key, val):
    return ''


def meta_end_s():
    return ''


def type_begin_s(tname, ttype, headers, cls):
    assert len(headers) == len(cls)
    return ''


def type_item_s(row, cls):
    assert len(row) == len(cls)
    return ''


def type_end_s():
    return ''


# ---------- JADN-IDL output ------------------


def doc_begin_d(title):
    text = ''
    return text


def doc_end_d():
    return ''


def sect_d(num, name):
    return ''


def meta_begin_d():
    return ''


def meta_item_d(key, val):
    return ''


def meta_end_d():
    return ''


def type_begin_d(tname, ttype, headers, cls):
    assert len(headers) == len(cls)
    return tname + ' = ' + ttype + '\n' if ttype else ''


def type_item_d(row, cls):
    assert len(row) == len(cls)
    return ''


def type_end_d():
    return ''


# ---------- CDDL output ------------------


def doc_begin_c(title):
    text = ''
    return text


def doc_end_c():
    return ''


def sect_c(num, name):
    return ''


def meta_begin_c():
    return ''


def meta_item_c(key, val):
    return ''


def meta_end_c():
    return ''


def type_begin_c(tname, ttype, headers, cls):
    assert len(headers) == len(cls)
    return ''


def type_item_c(row, cls):
    assert len(row) == len(cls)
    return ''


def type_end_c():
    return ''


# ---------- Thrift output ------------------


def doc_begin_t(title):
    text = ''
    return text


def doc_end_t():
    return ''


def sect_t(num, name):
    return ''


def meta_begin_t():
    return ''


def meta_item_t(key, val):
    return ''


def meta_end_t():
    return ''


def type_begin_t(tname, ttype, headers, cls):
    assert len(headers) == len(cls)
    return ''


def type_item_t(row, cls):
    assert len(row) == len(cls)
    return ''


def type_end_t():
    return ''

# ----------------------------------------------

"""
doc_begin - initial content
doc_end - closing content, if any
sect - section heading for human document formats, nothing for machine-readable schemas
meta_begin - begin meta content
meta_item - most meta items
meta_imps - special handling for imports meta item
meta_end - close meta content
type_begin - begin type definition
type_item - add field to type definition
type_end - close type definition
"""
# TODO: refactor into base and sub classes to support instance context

wtab = {
    'jas': (doc_begin_s, doc_end_s, sect_s, meta_begin_s, meta_item_s, meta_end_s, type_begin_s, type_item_s, type_end_s),
    'jidl': (doc_begin_d, doc_end_d, sect_d, meta_begin_d, meta_item_d, meta_end_d, type_begin_d, type_item_d, type_end_d),
    'cddl': (doc_begin_c, doc_end_c, sect_c, meta_begin_c, meta_item_c, meta_end_c, type_begin_c, type_item_c, type_end_c),
    'html': (doc_begin_h, doc_end_h, sect_h, meta_begin_h, meta_item_h, meta_end_h, type_begin_h, type_item_h, type_end_h),
    'thrift': (doc_begin_t, doc_end_t, sect_t, meta_begin_t, meta_item_t, meta_end_t, type_begin_t, type_item_t, type_end_t),
    'markdown': (doc_begin_m, doc_end_m, sect_m, meta_begin_m, meta_item_m, meta_end_m, type_begin_m, type_item_m, type_end_m)
}

DEFAULT_FORMAT = 'html'


def table_dumps(jadn, form=DEFAULT_FORMAT):
    """
    Translate JADN schema into other formats

    Column classes for presentation formats:
    n - number (right aligned)
    h - meta header (bold, right aligned)
    s - string (left aligned)
    b - bold (bold, left aligned)
    d - description (left aligned, extra width)
    """

    def _tbegin(to, name, tdef, head, cls):
        h = head
        c = cls
        if 'id' in to:
            h = [head[0]] + head[2:]
            c = [cls[0]] + cls[2:]
        return type_begin(name, tdef, h, c)

    def _titem(to, fitems, cls):
        f = fitems
        c = cls
        if 'id' in to:
            f = [fitems[0]] + fitems[2:]
            label = '**' + fitems[1] + '**::' if fitems[1] else ''
            f[-1] = label + f[-1]
            c = [cls[0]] + cls[2:]
        return type_item(f, c)

    def _fieldstr(typestr, opts):
        if 'tfield' in opts:
            return typestr + '(&' + opts['tfield'] + ')'
        return typestr

    doc_begin, doc_end, sect, meta_begin, meta_item, meta_end, type_begin, type_item, type_end = wtab[form]
    meta = jadn['meta']
    config = get_config(meta)
    title = meta['module'] + (' v.' + meta['patch']) if 'patch' in meta else ''
    text = doc_begin(title)
    text += meta_begin()
    meta_list = ('title', 'module', 'patch', 'description', 'exports', 'imports', 'bounds')
    bn = ('max_msg', 'max_str', 'max_bin', 'max_fields')
    for h in meta_list + tuple(set(meta) - set(meta_list)):
        if h in meta:
            mh = zip(bn, meta[h]) if h == 'bounds' else meta[h]
            text += meta_item(h, mh)
    text += meta_end()

    for td in jadn['types']:
        to = topts_s2d(td[TypeOptions])
        ttype = typestring(td[BaseType], to)
        if not is_builtin(td[BaseType]):
            text += 'Error - bad type: ' + str(td) + '\n'
        elif td[BaseType] in (SIMPLE_TYPES + ('ArrayOf', 'MapOf')) or 'enum' in to or 'pointer' in to:
            cls = ['b', 's', 'd']
            text += type_begin('', None, ['Type Name', 'Type Definition', 'Description'], cls)
            text += type_item([td[TypeName], ttype, td[TypeDesc]], cls)
        elif td[BaseType] == 'Enumerated':
            cls = ['n', 'b', 'd']
            text += _tbegin(to, td[TypeName], ttype, ['ID', 'Name', 'Description'], cls)
            for fd in td[Fields]:
                text += _titem(to, [str(fd[ItemID]), fd[ItemValue], fd[ItemDesc]], cls)
        else:                                       # Array, Choice, Map, Record
            cls = ['n', 'b', 's', 'n', 'd']
            if td[BaseType] == 'Array':
                cls2 = ['n', 's', 'n', 'd']         # Don't print ".ID" in type name but display fields as compact
                text += _tbegin(to, td[TypeName], ttype, ['ID', 'Type', '#', 'Description'], cls2)
                to.update({'id': True})
            else:
                text += _tbegin(to, td[TypeName], ttype, ['ID', 'Name', 'Type', '#', 'Description'], cls)
            for fd in td[Fields]:
                ft, fto = ftopts_s2d(fd[FieldOptions])
                fo = {'minc': 1, 'maxc': 1}
                fo.update(ft)
                ftype = _fieldstr(typestring(fd[FieldType], fto), fo)
                fn = fd[FieldName] + ('/' if 'dir' in fo else '')
                text += _titem(to, [str(fd[FieldID]), fn, ftype, multiplicity(fo['minc'], fo['maxc']), fd[FieldDesc]], cls)
        text += type_end()

    text += doc_end()
    return text


def table_dump(jadn, fname, source='', form=DEFAULT_FORMAT):
    with open(fname, 'w', encoding='utf8') as f:
        if source:
            f.write('<!-- Generated from ' + source + ', ' + datetime.ctime(datetime.now()) + '-->\n')
        f.write(table_dumps(jadn, form))



"""
Load, validate, prettyprint, and dump JSON Abstract Encoding Notation (JADN) schemas
"""

import json
import jsonschema
import numbers
import os
from datetime import datetime

from jadn.utils import topts_s2d, ftopts_s2d, build_deps, raise_error
from jadn.codec import Codec
from jadn.definitions import *


def schema_dir():
    """
    Return directory containing JADN schema files
    """
    return os.path.abspath(os.path.dirname(__file__))


def check(schema):
    """
    Validate JADN schema against JSON schema,
    Validate JADN schema against JADN meta-schema, then
    Perform additional checks on type definitions
    """

    def check_typeopts(type_name, base_type, topts):  # Check for invalid type options and undefined formats
        ro = {k for k in REQUIRED_TYPE_OPTIONS[base_type]} - {k for k in topts}
        if ro:
            raise_error('Missing type option', type_name + ':', str(ro))
        uo = {k for k in topts} - {k for k in ALLOWED_TYPE_OPTIONS[base_type]}
        if uo:
            raise_error('Unsupported type option', type_name + ' (' + base_type + '):', str(uo))
        if 'format' in topts:
            f = topts['format']
            fm = dict(list(FORMAT_VALIDATE.items()) + list(FORMAT_JS_VALIDATE.items()) + list(FORMAT_SERIALIZE.items()))
            if f not in fm or base_type != fm[f]:
                raise_error('Unsupported format', f, 'in', type_name, base_type)
        if 'enum' in topts and 'pointer' in topts:
            raise_error('Type cannot be both Enum and Pointer', type_name, base_type)
        if 'and' in topts and 'or' in topts:
            raise_error('Unsupported union+intersection in ', type_name, base_type)

    here = schema_dir()
    with open(os.path.join(here, 'jadn_schema.json')) as f:       # Check using JSON Schema for JADN
        jsonschema.Draft7Validator(json.load(f)).validate(schema)

    with open(os.path.join(here, 'jadn_schema.jadn')) as f:       # Optional: check using JADN meta-schema
        meta_schema = Codec(json.load(f), verbose_rec=True, verbose_str=True, config=schema)
        assert meta_schema.encode('Schema', schema) == schema

    types = set()                                                   # Additional checks not included in schema
    for t in schema['types']:
        tn = t[TypeName]
        bt = t[BaseType]
        if tn in types:
            raise_error('Duplicate type definition', tn)
        types |= {tn}
        if is_builtin(tn):
            raise_error('Reserved type name', tn)
        if not is_builtin(bt):
            raise_error('Invalid base type', tn + ':', bt)
        topts = topts_s2d(t[TypeOptions])
        check_typeopts(tn, bt, topts)
        flen = 0 if 'enum' in topts or 'pointer' in topts else FIELD_LENGTH[bt]
        if flen and len(t) <= Fields:
            raise_error('Missing fields', tn + '(' + bt + ')')
        elif not flen and len(t) > Fields:
            raise_error(tn + '(' + bt + ')', 'Cannot have fields')

        if flen:                # Check fields
            fids = set()        # Field IDs
            fnames = set()      # Field Names
            ordinal = bt in ('Array', 'Record')
            for n, f in enumerate(t[Fields]):
                if len(f) != flen:
                    raise_error('Bad field', n + 1, 'in', tn, 'length', len(f), 'should be', flen)
                fids.update({f[FieldID]})
                fnames.update({f[FieldName]})
                if ordinal and f[FieldID] != n + 1:
                    raise_error('Item tag error:', tn + '(' + bt + ')[' + f[FieldName] + '] --', f[FieldID], 'should be', n + 1)
                if flen > FieldDesc:    # Full field, not an Enumerated item
                    fo, fto = ftopts_s2d(f[FieldOptions])
                    if is_builtin(f[FieldType]):
                        check_typeopts(tn + '/' + f[FieldName], f[FieldType], fto)
                        if 'minc' in fo and 'maxc' in fo:
                            if fo['minc'] < 0 or (fo['maxc'] > 0 and fo['maxc'] < fo['minc']):
                                raise_error(tn + '/', f[FieldName], 'bad cardinality', fo['minc'], fo['maxc'])
                    elif fto:
                        # raise_error(tn + '/' + f[FieldName] + '(' + f[FieldType] + ') cannot have Type options', fto)
                        pass    # TODO: pass 'unique' to multiplicity-based ArrayOf
            if len(t[Fields]) != len(fids) or len(t[Fields]) != len(fnames):
                raise_error('Duplicate field', tn, len(t[Fields]), 'fields,', len(fids), 'unique tags', len(fnames), 'unique names')
    return schema


def analyze(schema):
    def ns(name, nsids):   # Return namespace if name has a known namespace, otherwise return full name
        nsp = name.split(':')[0]
        return nsp if nsp in nsids else name

    items = build_deps(schema)
    # out, roots = topo_sort(items)
    imports = schema['meta']['imports'] if 'imports' in schema['meta'] else {}
    exports = schema['meta']['exports'] if 'exports' in schema['meta'] else []
    defs = {i for i in items} | set(imports)
    refs = set().union([ns(r, imports) for i in items for r in items[i]]) | set(exports)
    oids = (OPTION_ID['enum'], OPTION_ID['pointer'])
    refs = {r[1:] if r[0] in oids else r for r in refs}         # Reference base type for derived enums/pointers
    return {
        'unreferenced': [str(k) for k in defs - refs],
        'undefined': [str(k) for k in refs - defs],
        'cycles': [],
    }


def loads(jadn_str):
    schema = json.loads(jadn_str)
    return check(schema)


def load(fname):
    with open(fname, encoding='utf-8') as f:
        schema = json.load(f)
    return check(schema)


def dumps(schema, level=0, indent=1, strip=False, nlevel=None):
    sp = level * indent * ' '
    sp2 = (level + 1) * indent * ' '
    sep2 = ',\n' if strip else ',\n\n'
    if isinstance(schema, dict):
        sep = ',\n' if level > 0 else sep2
        lines = []
        for k in schema:
            lines.append(sp2 + '"' + k + '": ' + dumps(schema[k], level + 1, indent, strip))
        return '{\n' + sep.join(lines) + '\n' + sp + '}'
    elif isinstance(schema, list):
        sep = ',\n' if level > 1 else sep2
        vals = []
        nest = schema and isinstance(schema[0], list)       # Not an empty list
        for v in schema:
            sp3 = sp2 if nest else ''
            vals.append(sp3 + dumps(v, level + 1, indent, strip, level))
        if nest:
            spn = (nlevel if nlevel else level) * indent * ' '
            return '[\n' + sep.join(vals) + '\n' + spn + ']'
        return '[' + ', '.join(vals) + ']'
    elif isinstance(schema, (numbers.Number, type(''))):
        return json.dumps(schema)
    return '???'


def dump(schema, fname, source='', strip=False):
    with open(fname, 'w') as f:
        if source:
            f.write('"Generated from ' + source + ', ' + datetime.ctime(datetime.now()) + '"\n\n')
        f.write(dumps(schema, strip=strip) + '\n')




"""
 JADN Definitions

A JSON Abstract Data Notation (JADN) file contains a list of datatype definitions.  Each type definition
has a specified format - a list of four or five columns depending on whether the type is primitive or
structure: (name, base type, type options, type description [, fields]).

For the enumerated type each field definition is a list of three items: (tag, name, description).

For other structure types (array, choice, map, record) each field definition is a list of five items:
(tag, name, type, field options, field description).
"""

# Datatype Definition columns
TypeName = 0            # Name of the type being defined
BaseType = 1            # Core type of the type being defined
TypeOptions = 2         # An array of zero or more TYPE_OPTIONS
TypeDesc = 3            # A non-normative description of the type
Fields = 4              # List of one or more items or fields

# Enumerated Item Definition columns
ItemID = 0              # Integer item identifier
ItemValue = 1           # String value of the item
ItemDesc = 2            # A non-normative description of the Enumerated item

# Field Definition columns
FieldID = 0             # Integer field identifier
FieldName = 1           # Name or label of the field
FieldType = 2           # Type of the field
FieldOptions = 3        # An array of zero or more FIELD_OPTIONS (and TYPE_OPTIONS if extended)
FieldDesc = 4           # A non-normative description of the field

# Core datatypes

SIMPLE_TYPES = (
    'Binary',
    'Boolean',
    'Integer',
    'Number',
    'Null',
    'String',
)

SELECTOR_TYPES = (
    'Enumerated',       # enum option specifies fields derived from a defined type
    'Choice',
)

CONTAINER_TYPES = (
    'Array',
    'ArrayOf',          # (value_type): instance is a container but definition has no fields
    'Map',
    'MapOf',            # (key_type, value_type): instance is a container but definition has no fields
    'Record',
)

CORE_TYPES = SIMPLE_TYPES + SELECTOR_TYPES + CONTAINER_TYPES

FIELD_LENGTH = {
    'Binary': 0,
    'Boolean': 0,
    'Integer': 0,
    'Number': 0,
    'Null': 0,
    'String': 0,
    'Enumerated': 3,    # 0 if Enumerated type definition contains enum or pointer option
    'Choice': 5,
    'Array': 5,
    'ArrayOf': 0,
    'Map': 5,
    'MapOf': 0,
    'Record': 5,
}


def is_builtin(t):      # Is a core type
    return t in CORE_TYPES


def has_fields(t):      # Is a type with fields listed in definition
    return FIELD_LENGTH[t] == 5 if is_builtin(t) else False


# Option Tags/Keys
#   JADN TypeOptions and FieldOptions contain a list of strings, each of which is an option.
#   The first character of an option string is the type ID; the remaining characters are the value.
#   The option string is converted into a Name: Value pair before use.
#   The tables list the unicode codepoint of the ID and the corresponding Name.

TYPE_OPTIONS = {        # ID, value type, description
    0x3d: 'id',         # '=', none, Enumerated type and Choice/Map/Record keys are ID not Name
    0x2a: 'vtype',      # '*', string, Value type for ArrayOf and MapOf
    0x2b: 'ktype',      # '+', string, Key type for MapOf
    0x23: 'enum',       # '#', string, enumeration derived from the referenced Array/Choice/Map/Record type
    0x3e: 'pointer',    # '>', string, enumeration of pointers derived from the referenced Array/Choice/Map/Record type
    0x2f: 'format',     # '/', string, semantic validation keyword, may affect serialization
    0x25: 'pattern',    # '%', string, regular expression that a string must match
    0x7b: 'minv',       # '{', integer, minimum byte or text string length, numeric value, element count
    0x7d: 'maxv',       # '}', integer, maximum byte or text string length, numeric value, element count
    0x71: 'unique',     # 'q', none, ArrayOf instance must not contain duplicates
    0x2229: 'and',      # '∩', string, INTERSECTION - instance must also match referenced type definition (allOf)
    0x222a: 'or',       # '∪', string, UNION - instance must match at least one of the type definitions (anyOf)
}

FIELD_OPTIONS = {
    0x5b: 'minc',       # '[', integer, minimum cardinality, default = 1, 0 = field is optional
    0x5d: 'maxc',       # ']', integer, maximum cardinality, default = 1, 0 = inherited max, not 1 = array
    0x26: 'tfield',     # '&', string, field that specifies the type of this field
    0x3c: 'dir',        # '<', none, pointer enumeration uses field name as path component (field is not a leaf node)
    0x21: 'default',    # '!', string, default value for an instance of this type
}

OPTION_ID = {           # Reverse index - must match TYPE_OPTIONS and FIELD_OPTIONS
    'id':       chr(0x3d),
    'vtype':    chr(0x2a),
    'ktype':    chr(0x2b),
    'enum':     chr(0x23),
    'pointer':  chr(0x3e),
    'format':   chr(0x2f),
    'pattern':  chr(0x25),
    'minv':     chr(0x7b),
    'maxv':     chr(0x7d),
    'unique':   chr(0x71),
    'and':      chr(0x2229),
    'or':       chr(0x222a),
    'minc':     chr(0x5b),
    'maxc':     chr(0x5d),
    'tfield':   chr(0x26),
    'dir':      chr(0x3c),
    'default':  chr(0x21),
}

REQUIRED_TYPE_OPTIONS = {
    'Binary': [],
    'Boolean': [],
    'Integer': [],
    'Number': [],
    'Null': [],
    'String': [],
    'Enumerated': [],
    'Choice': [],
    'Array': [],
    'ArrayOf': ['vtype'],
    'Map': [],
    'MapOf': ['ktype', 'vtype'],
    'Record': [],
}

ALLOWED_TYPE_OPTIONS = {
    'Binary': ['and', 'or', 'minv', 'maxv', 'format'],
    'Boolean': ['and', 'or'],
    'Integer': ['and', 'or', 'minv', 'maxv', 'format'],
    'Number': ['and', 'or', 'minv', 'maxv', 'format'],
    'Null': ['and', 'or'],
    'String': ['and', 'or', 'minv', 'maxv', 'format', 'pattern'],
    'Enumerated': ['and', 'or', 'id', 'enum', 'pointer'],
    'Choice': ['and', 'or', 'id'],
    'Array': ['and', 'or', 'minv', 'maxv', 'format'],
    'ArrayOf': ['and', 'or', 'vtype', 'minv', 'maxv', 'unique'],
    'Map': ['and', 'or', 'id', 'minv', 'maxv'],
    'MapOf': ['and', 'or', 'ktype', 'vtype', 'minv', 'maxv'],
    'Record': ['and', 'or', 'minv', 'maxv'],
}

FORMAT_JS_VALIDATE = {      # Semantic validation formats defined by JSON Schema 2019-09 Sec 7.3
    'date-time': 'String',
    'date': 'String',
    'time': 'String',
    'duration': 'String',
    # 'email': 'String',        # Python jsonschema package has deliberately buggy email - won't be fixed
    'idn-email': 'String',
    # 'hostname': 'String',     # jsonschema package needs bug fix
    'idn-hostname': 'String',
    'ipv4': 'String',           # doesn't allow netmask prefix length
    'ipv6': 'String',           # doesn't allow netmask prefix length
    'uri': 'String',
    'uri-reference': 'String',
    'iri': 'String',
    'iri-reference': 'String',
    'uuid': 'String',
    'uri-template': 'String',
    'json-pointer': 'String',
    'relative-json-pointer': 'String',
    'regex': 'String'
}

FORMAT_VALIDATE = {         # Semantic validation formats defined by JADN
    'email': 'String',          # Use this instead of jsonschema
    'hostname': 'String',       # Use this instead of jsonschema
    'eui': 'Binary',            # IEEE Extended Unique Identifier, 48 bits or 64 bits
    'ipv4-addr': 'Binary',      # IPv4 address as specified in RFC 791 Section 3.1
    'ipv6-addr': 'Binary',      # IPv6 address as specified in RFC 8200 Section 3
    'ipv4-net': 'Array',        # Binary IPv4 address and Integer prefix length, RFC 4632 Section 3.1
    'ipv6-net': 'Array',        # Binary IPv6 address and Integer prefix length, RFC 4291 Section 2.3
    'i8': 'Integer',            # Signed 8 bit integer [-128 .. 127]
    'i16': 'Integer',           # Signed 16 bit integer [-32768 .. 32767]
    'i32': 'Integer',           # Signed 32 bit integer [-2147483648 .. 2147483647]
    'i64': 'Integer',           # Signed 64 bit integer [-2^63 .. 2^63 -1]
    # 'u#': 'Integer',            # Unsigned "#"-bit integer or bit field where #>0, [0 .. 2^# -1]
}

FORMAT_SERIALIZE = {        # Data representation formats for one or more serializations
    'eui': 'Binary',            # IEEE EUI, "hex-byte-colon" text representation, (e.g., 00:1B:44:11:3A:B7)
    'ipv4-addr': 'Binary',      # IPv4 "dotted-quad" text representation, RFC 2673 Section 3.2
    'ipv6-addr': 'Binary',      # IPv6 text representation, RFC 4291 Section 2.2
    'ipv4-net': 'Array',        # IPv4 Network Address CIDR text string, RFC 4632 Section 3.1
    'ipv6-net': 'Array',        # IPv6 Network Address CIDR text string, RFC 4291 Section 2.3
    'b': 'Binary',              # Base64url - RFC 4648 Section 5 (default text representation of Binary type)
    'x': 'Binary',              # Hex - RFC 4648 Section 8
    'i8': 'Integer',            # 8 bit field - these affect packed (RFC 791 style) serializations
    'i16': 'Integer',           # 16 bit field
    'i32': 'Integer',           # 32 bit field
    'i64': 'Integer',           # 64 bit field
    # 'u#': 'Integer',            # #-bit field
    'f16': 'Number',            # IEEE 754 Half-Precision Float - these affect CBOR serialization
    'f32': 'Number',            # IEEE 754 Single-Precision Float
    'f64': 'Number',            # IEEE 754 Double-Precision Float (default binary representation of Number type)
}

DEFAULT_CONFIG = {          # Configuration values to use if not specified in schema
    "$MaxBinary": 255,          # Maximum number of octets for Binary types
    "$MaxString": 255,          # Maximum number of characters for String types
    "$MaxElements": 100,        # Maximum number of items/properties for container types
    "$Sys": "$",                # System reserved character for TypeName
    "$TypeName": "^[A-Z][-$A-Za-z0-9]{0,31}$",   # Type Name regex
    "$FieldName": "^[a-z][_A-Za-z0-9]{0,31}$",   # Field Name regex
    "$NSID": "^[A-Za-z][A-Za-z0-9]{0,7}$",       # Namespace ID regex
    "$TypeRef": "^$"            # Placeholder.  Actual pattern is ($NSID ":")? $TypeName
}

JADN_EXTENSIONS = {
    'AnonymousType',            # TYPE_OPTIONS included in FieldOptions
    'Multiplicity',             # maxc other than 1, or minv other than 0 (optional) or 1 (required)
    'DerivedEnum',              # enum and pointer/dir options, create Enumerated type of fields or JSON Pointers
    'MapOfEnum',                # ktype option specifies an Enumerated type
}


from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='jadn',
    version='0.5.1',
    description='JADN schema tools',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/davaya/jadn-pypkg',
    author='David Kemp',
    author_email='dk190a@gmail.com',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
    ],
    keywords='schema',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    python_requires='>=3.4',
    install_requires=['jsonschema'],
    package_data={
        'jadn': ['jadn_schema.jadn', 'jadn_schema.json']
    }
)


import base64
import binascii
import socket
from socket import AF_INET, AF_INET6
import string
from jadn.definitions import *

"""
Create a table listing the serialization functions for each format keyword
"""


def json_format_codecs():                              # return table of JSON format serialization functions
    # Ensure code is in sync with JADN definitions
    assert set(FORMAT_SERIALIZE) == \
           set(FORMAT_CONVERT_BINARY_FUNCTIONS) |\
           set(FORMAT_CONVERT_MULTIPART_FUNCTIONS) |\
           set(FORMAT_CONVERT_INTEGER_FUNCTIONS) |\
           set(FORMAT_CONVERT_NUMBER_FUNCTIONS)

    return {
        'Binary': FORMAT_CONVERT_BINARY_FUNCTIONS,
        'Array': FORMAT_CONVERT_MULTIPART_FUNCTIONS,
        'Integer': FORMAT_CONVERT_INTEGER_FUNCTIONS,
        'Number': FORMAT_CONVERT_NUMBER_FUNCTIONS
    }


"""
Return serialization functions for the specified format keyword and data type
"""


def get_format_encode_function(format_table, base_type, format_kw):
    return _codec_function(format_table, base_type, format_kw, 0)


def get_format_decode_function(format_table, base_type, format_kw):
    return _codec_function(format_table, base_type, format_kw, 1)


def _codec_function(format_table, base_type, format_kw, direction):
    try:
        if not format_kw:
            format_kw = {'Binary': 'b', 'Number': 'f64'}[base_type]   # Set default format for type if one exists
        return format_table[base_type][format_kw][direction]
    except KeyError:
        return _format_pass


def _format_pass(val):
    return val


"""
Binary to String, String to Binary conversion functions
"""


def b2s_hex(bval):      # Convert from binary to hex string
    return base64.b16encode(bval).decode()


def s2b_hex(sval):      # Convert from hex string to binary
    try:
        return base64.b16decode(sval)
    except binascii.Error:
        raise TypeError


def b2s_base64url(bval):      # Convert from binary to base64url string
    return base64.urlsafe_b64encode(bval).decode().rstrip('=')


def s2b_base64url(sval):      # Convert from base64url string to binary
    v = sval + ((4 - len(sval) % 4) % 4)*'='          # Pad b64 string out to a multiple of 4 characters
    if set(v) - set(string.ascii_letters + string.digits + '-_='):  # Python 2 doesn't support Validate
        raise TypeError('base64decode: bad character')
    return base64.b64decode(str(v), altchars='-_')


def b2s_ipv4_addr(bval):      # Convert IPv4 address from binary to string
    try:
        return socket.inet_ntop(AF_INET, bval)
    except AttributeError:
        try:
            return socket.inet_ntoa(bval)       # Python 2 doesn't support inet_ntop on Windows
        except IOError:
            raise ValueError
    except OSError:
        raise ValueError


def b2s_ipv6_addr(bval):        # Convert ipv6 address from binary to string
    try:
        return socket.inet_ntop(AF_INET6, bval)     # Python 2 doesn't support inet_ntop on Windows
    except OSError:
        raise ValueError


def s2b_ipv4_addr(sval):    # Convert IPv4 addr from string to binary
    try:
        return socket.inet_pton(AF_INET, sval)
    except AttributeError:       # Python 2 doesn't support inet_pton on Windows
        try:
            return socket.inet_aton(sval)
        except IOError:
            raise ValueError
    except OSError:
        raise ValueError


def s2b_ipv6_addr(sval):    # Convert IPv6 address from string to binary
    try:
        return socket.inet_pton(AF_INET6, sval)
    except OSError:         # Python 2 doesn't support inet_pton on Windows
        raise ValueError


FORMAT_CONVERT_BINARY_FUNCTIONS = {
    'b': (b2s_base64url, s2b_base64url),            # Base64url
    'x': (b2s_hex, s2b_hex),                        # Hex
    'ipv4-addr': (b2s_ipv4_addr, s2b_ipv4_addr),    # IPv4 Address
    'ipv6-addr': (b2s_ipv6_addr, s2b_ipv6_addr),    # IPv6 Address
    'eui': (b2s_hex, s2b_hex),                      # EUI - TODO: write colon-hex A0:32:F9:...
}


"""
IP Net (address, prefix length tuple) conversions
"""


def a2s_ipv4_net(aval):
    if aval[1] < 0 or aval[1] > 32:         # Verify prefix length is valid
        raise ValueError
    sa = b2s_ipv4_addr(aval[0])             # Convert Binary bytes to type-specific string
    return sa + '/' + str(aval[1])


def a2s_ipv6_net(aval):
    if aval[1] < 0 or aval[1] > 128:        # Verify prefix length is valid
        raise ValueError
    sa = b2s_ipv6_addr(aval[0])             # Convert Binary bytes to type-specific string
    return sa + '/' + str(aval[1])


def s2a_ipv4_net(sval):
    sa, spl = sval.split('/', 1)
    sa = s2b_ipv4_addr(sa)                  # Convert type-specific string to Binary bytes
    prefix_len = int(spl)
    if prefix_len < 0 or prefix_len > 32:
        raise ValueError
    return [sa, prefix_len]


def s2a_ipv6_net(sval):
    sa, spl = sval.split('/', 1)
    sa = s2b_ipv6_addr(sa)                  # Convert type-specific string to Binary bytes
    prefix_len = int(spl)
    if prefix_len < 0 or prefix_len > 128:
        raise ValueError
    return [sa, prefix_len]


FORMAT_CONVERT_MULTIPART_FUNCTIONS = {
    'ipv4-net': (a2s_ipv4_net, s2a_ipv4_net),       # IPv4 Net Address with CIDR prefix length
    'ipv6-net': (a2s_ipv6_net, s2a_ipv6_net),       # IPv6 Net Address with CIDR prefix length
}

# No special serialization in JSON.  Define these for packed encoding.
FORMAT_CONVERT_INTEGER_FUNCTIONS = {
    'i8': (_format_pass, _format_pass),
    'i16': (_format_pass, _format_pass),
    'i32': (_format_pass, _format_pass),
    'i64': (_format_pass, _format_pass)
}

# No special serialization in JSON. Define these for CBOR encoding.
FORMAT_CONVERT_NUMBER_FUNCTIONS = {
    'f16': (_format_pass, _format_pass),
    'f32': (_format_pass, _format_pass),
    'f64': (_format_pass, _format_pass)
}
